#! /usr/bin/env python

#############################################################################
##                                                                         ##
## ipv6world.py ---  IPv6 backbone explorer toy                            ##
##              see http://www.secdev.org/projects/                        ##
##              for more informations                                      ##
##                                                                         ##
## Copyright (C) 2007  Philippe Biondi <phil@secdev.org>                   ##
##                                                                         ##
## This program is free software; you can redistribute it and/or modify it ##
## under the terms of the GNU General Public License version 2 as          ##
## published by the Free Software Foundation.                              ##
##                                                                         ##
## This program is distributed in the hope that it will be useful, but     ##
## WITHOUT ANY WARRANTY; without even the implied warranty of              ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        ##
## General Public License for more details.                                ##
##                                                                         ##
#############################################################################


from scapy6_teredo import *
import logging,thread
from threading import Semaphore

log = logging.getLogger("probe_ipv6")
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter("P>%(levelname)-5s: %(message)s"))
log.addHandler(console_handler)
log.setLevel(2)

max_thread_sema = Semaphore(30)


def do_traceroute(*dst):
    firstdst = dst[0]
    waypoints = dst[1:]
    maxttl = min(255, 30*len(dst))
    if waypoints:
        res,unans= traceroute6(firstdst, maxttl=maxttl, l4=IPv6ExtHdrRouting(addresses=waypoints)/TCP(sport=RandShort(), dport=80), retry=-1, timeout=2, verbose=0)
    else:
        res,unans = traceroute6(firstdst,verbose=0, retry=-1, timeout=2)
    dct, = res.get_trace().values()
    trace = []
    i = 1
    mx = max(dct.keys())
    while i <= mx:
        if i in dct:
            ip,done = dct[i]
            trace.append(ip)
            if done:
                break
        else:
            trace.append("")
        i += 1
    return repr(trace)

def do_UDP_traceroute(*dst):
    firstdst = dst[0]
    waypoints = dst[1:]
    maxttl = min(255, 30*len(dst))
    if waypoints:
        res,unans= traceroute6(firstdst, maxttl=maxttl, l4=IPv6ExtHdrRouting(addresses=waypoints)/UDP(sport=RandShort(), dport=53), retry=-1, timeout=2, verbose=0)
    else:
        res,unans = traceroute6(firstdst,l4=UDP(sport=RandShort(), dport=53),verbose=0, retry=-1, timeout=2)
    dct, = res.get_trace().values()
    trace = []
    i = 1
    mx = max(dct.keys())
    while i <= mx:
        if i in dct:
            ip,done = dct[i]
            trace.append(ip)
            if done:
                break
        else:
            trace.append("")
        i += 1
    return repr(trace)


def do_reverse_dns(ip):
    try:
        ret = socket.gethostbyaddr(ip)[0]
    except socket.herror:
        log.info("Reverse DNS lookup for %s failed" % ip)
        ret = ""
    except socket.gaierror:
        log.warning("Reverse DNS lookup service not possible on %s" % ip)
        ret = ""
    return repr(ret)


def do_AS(ip):
    res = conf.AS_resolver.resolve(ip)
    if res:
        ret = res[0][1:]
    else:
        log.info("AS lookup for %s failed" % ip)
        ret = [0,""]
    return repr(ret)

def do_test_rh(ip):
    # F.ROOT-SERVERS.NET: 2001:500::1035
    res,unans = sr(IPv6(dst=ip)/IPv6ExtHdrRouting(addresses=["2001:500::1035"])/ICMPv6EchoRequest(),
                   retry=-2, timeout=3, verbose=0)
    return repr(1*(len(res) and ICMPv6EchoReply in res[0][1]))

def do_scan(ip):
    res,unans = sr(IPv6(dst=ip)/TCP(sport=RandShort(), dport=[21,22,23,25,53,80,110,143,179,443,2602,2604,8080]),
                   verbose=0, timeout=3, retry=-3)
    ret = []
    for s,r in res:
        if TCP in r and r[TCP].flags & 18 == 18:
            log.debug("Port open: %r" % r)
            ret.append(r[TCP].sport)
    return repr(ret)

def do_niq_ipv6(tgt):
    try:
        socket.inet_pton(socket.AF_INET6, tgt)
    except socket.error:
            tgt = socket.getaddrinfo(tgt, None, socket.AF_INET6)[0][-1][0]

    resp = sr1(IPv6(dst=tgt)/ICMPv6NIQueryIPv6(nonce=RandString(8), data=tgt), 
                   timeout=1, retry=-3, verbose=0)

    if (resp is not None and 
        ICMPv6NIReplyIPv6 in resp and 
        resp[ICMPv6NIReplyIPv6].code == 0):
        return repr(map(lambda x: x[1], resp[ICMPv6NIReplyIPv6].data))

    return "''"

def do_niq_name(tgt):
    try:
        socket.inet_pton(socket.AF_INET6, tgt)
    except socket.error:
            tgt = socket.getaddrinfo(tgt, None, socket.AF_INET6)[0][-1][0]

    resp = sr1(IPv6(dst=tgt)/ICMPv6NIQueryName(nonce=RandString(8), data=tgt), 
                   timeout=1, retry=-3, verbose=0)

    if (resp is not None and 
        ICMPv6NIReplyName in resp and 
        resp[ICMPv6NIReplyName].code == 0):
        return resp[ICMPv6NIReplyName].data[1]

    return "''"

def async_run(retfile, cid, func, args):
    max_thread_sema.acquire()
    
    def async_thread(retfile=retfile, cid=cid, func=func, args=args):
        try:
            try:
                res = func(*args)
            except:
                log.exception("Async thread %s (%s applied on %r)" % (cid,func.__name__, args))
            else:
                log.debug("[%s] ==> %r" % (cid,res))
                retfile.write("answer %s %s\n" % (cid, res))
                retfile.flush()
        finally:
            max_thread_sema.release()
            
    thread.start_new_thread(async_thread, ())


if __name__ == "__main__":
    import sys
    from tools import MultiInput
    
    input = MultiInput(sys.stdin)
    log.debug("Begin main loop")
    try:
        while 1:
            try:
                cmd = input.get_command().split()
                if not cmd:
                    continue
                if cmd[0] in ["exit", "quit"]:
                    raise SystemExit
                if len(cmd) < 2:
                    log.warning("Bad syntax %r. command Id is missing?" % cmd)
                    continue
                log.debug("Got command %r" % cmd)
                func = "do_%s" % cmd[1]
                if func not in globals():
                    log.warning("Function %s not found" % func)
                    continue
                log.debug("[%s] Calling %s on %r" % (cmd[0], func, cmd[2:]))
                async_run(sys.stdout, cmd[0], globals()[func], cmd[2:])
            except KeyboardInterrupt:
                raise
            except SystemExit:
                raise
            except:
                log.exception("Exception in main loop")
    except KeyboardInterrupt:
        log.info("Interrupted by user.")
    except SystemExit:
        log.info("Interrupted by command.")
        
        
                
        
        
        
        
    
    


    
