#! /usr/bin/env python

#############################################################################
##                                                                         ##
## ipv6world.py ---  IPv6 backbone explorer toy                            ##
##              see http://www.secdev.org/projects/ipv6world               ##
##              for more informations                                      ##
##                                                                         ##
## Copyright (C) 2007  Philippe Biondi <phil@secdev.org>                   ##
##                                                                         ##
## This program is free software; you can redistribute it and/or modify it ##
## under the terms of the GNU General Public License version 2 as          ##
## published by the Free Software Foundation.                              ##
##                                                                         ##
## This program is distributed in the hope that it will be useful, but     ##
## WITHOUT ANY WARRANTY; without even the implied warranty of              ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        ##
## General Public License for more details.                                ##
##                                                                         ##
#############################################################################


import os, sys
import subprocess
import thread,re
from pprint import pprint
from visual import *
from rtgraph3d import *
import logging
log = logging.getLogger("ipv6world")
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter("%(levelname)-5s: %(message)s"))
log.addHandler(console_handler)
log.setLevel(2)

from tools import MultiInput

DT = 0.1
ACCEL=5

def tag_generator(tag="tag%04i", start=0):
    while 1:
        yield tag % start
        start += 1

EXISTING_PROPERTIES = [ "reverse_dns", "scan", "test_rh", "niq_name", "niq_ipv6"] #"AS" ]
#EXISTING_PROPERTIES = [ ]

class DBquery:
    def __init__(self, dbnet, ipdb):
        if dbnet is not None:
            log.info("Opening net DB [%s]" % dbnet)
            self.con = sqlite.connect(dbnet)
            self.cur = self.con.cursor()
            self.color = {None:(0.5,0.5,0.5)}
        else:
            self.ip2attr = lambda x: {"color":randcol()}
        if ipdb is not None:
            log.info("Opening IP DB [%s]" % ipdb)
            self.ipcon = sqlite.connect(ipdb)
            self.ipcur = self.ipcon.cursor()

    def execsql(self, sql):
        self.ipcur.execute(sql)
        pprint(self.ipcur.fetchall()) 

    def get_properties(self, ip):
        try:
            ipnum = ip2num(ip)
        except:
            log.exception("DBquery.ser_properties: can't split ip [%s]" % ip)
            return { }
        self.ipcur.execute("""select identifier, value
                              from ip6
                                   left join router on ip6.router_id = router.id
                                   left join router_properties on router_properties.router_id = router.id
                                   left join router_property_name
                                     on router_properties.router_property_id = router_property_name.id
                              where ip_hi = ? and ip_lo = ? order by ip6.date desc""", ipnum)
        res = self.ipcur.fetchall()
        if not res:
            return {}
        res = [(a.encode("ascii"),v.encode("ascii")) for a,v in res if a and v]
        log.debug("#####[%r]" % res)
        return dict(res)

    def get_router_id(self, ipnum):
        self.ipcur.execute("""select router.id
                              from ip6
                                   left join router on router.id = ip6.router_id
                              where ip_hi = ? and ip_lo = ? order by ip6.date desc""", ipnum)
        return self.ipcur.fetchone()[0]

    def set_prop(self, ipnum, propid, val):
        rid = self.get_router_id(ipnum)
        self.ipcur.execute("""insert into router_properties values (NULL,?,?,?,?)""",
                           (rid, propid, time.time(), val))
        self.ipcon.commit()
    
    def set_properties(self, ip, **props):
        try:
            ipnum = ip2num(ip)
        except:
            log.exception("DBquery.ser_properties: can't split ip [%s]" % ip)
            return
        
        self.ipcur.execute("""select id from router_property_name""")
        log.debug("}}}all props[%r]" % self.ipcur.fetchall())

        for prop,val in props.iteritems():            
            while 1:
                self.ipcur.execute("""select id from router_property_name where identifier=?""", (prop,))
                res = self.ipcur.fetchone()
                log.debug("set_properties: [%s] =(db)=> %r" % (prop, res))
                if res:
                    break
                self.ipcur.execute("""insert into router_property_name values (NULL,?,?)""", (prop, prop))
                self.ipcon.commit()
            propid = res[0]
            self.set_prop(ipnum, propid, val)

    def ip2attr(self, ip):
        try:
            ipnum = ip2num(ip)
        except:
            return { "color": self.color[None] }
        log.info("Query DB for IP [%s]" % ip)
        self.cur.execute("""select netname,country.name,country.a2
                            from inet6num
                                 left outer join inetinfo on inetinfo_id=inetinfo.id
                                 left outer join country on country_id=country.id
                            where prefix_hi = ? & mask_hi and
                                  prefix_lo = ? & mask_lo
                            order by length desc""", ipnum)
        res = self.cur.fetchone()
        if not res:
            return {}
        res = [x and x.encode("ascii") or "" for x in res]
        if res is None:
            log.error("IP [%s] not found" % ip)
            res = "??","??",None
        netname,country,a2 = res
        if a2 not in self.color:
            self.color[a2] = mkcol()
        attr = {"netname":netname, "country":country, "color":self.color[a2]}
        attr.update(self.get_properties(ip))
        return attr

    def add_ip(self, ip):
        self.ipcur.execute("""select count(*) from router where name = ?""", (ip,))
        if self.ipcur.fetchone()[0] > 0:
            return
        try:
            ipnum = ip2num(ip)
        except:
            return
        self.ipcur.execute("""insert into router values (NULL,?)""", (ip,))
        self.ipcon.commit()
        self.ipcur.execute("""select id from router where name = ?""", (ip,))
        rid = self.ipcur.fetchone()[0]
        self.ipcur.execute("""insert into ip6 values (NULL,?,?,?,?)""", (rid, ipnum[0], ipnum[1], time.time()))
        self.ipcon.commit()

    def add_link(self, ip1, ip2):
        try:
            ipnum1 = ip2num(ip1)
            ipnum2 = ip2num(ip2)
        except:
            return
        rid1 = self.get_router_id(ipnum1)
        rid2 = self.get_router_id(ipnum2)
        self.ipcur.execute("""select count(*) from link
                              where router1_id=? and router2_id=?
                                    or router1_id=? and router2_id=?""", (rid1,rid2,rid2,rid1))
        if self.ipcur.fetchone()[0] > 0:
            return
        self.ipcur.execute("""insert into link values (NULL, ?, ?, ?)""", (rid1, rid2, time.time()))
        self.ipcon.commit()

    def get_all_links(self):
        self.ipcur.execute("""select r1.name, r2.name
                              from link
                                   left join router as r1 on r1.id = link.router1_id
                                   left join router as r2 on r2.id = link.router2_id""")
        return self.ipcur.fetchall()
        

class GlowingService:
    def __init__(self):
        self.lst = {}
        self.run=1
        thread.start_new_thread(self.glowing_thread,())

    def __del__(self):
        self.run=0

    def glow_node(self, node):
        if node not in self.lst:
            self.lst[node]=(node.radius, node.color)

    def unglow_node(self, node):
        if node in self.lst:
            node.radius,node.color = self.lst.pop(node)

    def toggle_glow(self, node):
        if node in self.lst:
            self.unglow_node(node)
        else:
            self.glow_node(node)

    def glowing_thread(self):
        t=0
        while self.run:
            rate(25)
            mult = 1.1+0.3*cos(t*2*pi/25)
            for n,v in self.lst.iteritems():
                rad,col = v
                n.radius = rad*mult
                n.red = min(1,col[0]*mult)
                n.green = min(1,col[1]*mult)
                n.blue = min(1,col[2]*mult)
            t += 1

    def __iter__(self):
        return iter(self.lst)
                
            
class AsyncAnswer:
    def __init__(self):
        self.cb = {}
    def add_callback(self, cid, func):
        self.cb[cid] = func
    def callback(self, cid, parm):
        func = self.cb.pop(cid)
        try:
            func(*parm)
        except:
            log.exception("Callback %s:" % cid)
    


class Probe:
    def __init__(self, probecmd):
        self.probe = subprocess.Popen(probecmd,
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE,
#                                      stderr=subprocess.PIPE,
                                      close_fds=True)
        self.cmd_id = tag_generator("CMD%i")
        self.async_answer = AsyncAnswer()

    def send(self, cmd):
        if not cmd.endswith("\n"):
            cmd += "\n"
        self.probe.stdin.write(cmd)
        self.probe.stdin.flush()

    def send_with_cb(self, cmd, func):
        cid = self.cmd_id.next()
        self.async_answer.add_callback(cid, func)
        self.send("%s %s" % (cid,cmd))
        
    def parse_answer(self, cmd):
        _,cid,ans = cmd.split(" ",2)
        self.async_answer.callback(cid, [ans])
        
    def fileno(self):
        return self.probe.stdout.fileno()


def cinematic_thread(the_world, cmdpipe, ACCEL, DT):
    log.info("Cinematic thread started")
    scene = the_world.scene
    picked = None
    waypoints = []
    while 1:
        try:
            rate(ACCEL/DT)
            if scene.kb.keys:
                k = scene.kb.getkey()
                if k == "Q":
                    cmdpipe.write("quit\r")
                    return
            if scene.mouse.events:
                ev = scene.mouse.getevent()
                if ev.release == "left":
                    if picked:
                        the_world.speeds[picked.num]=((scene.mouse.pos-picked.PV)/DT).astuple()
                        the_world.cinematic_change = 1
                        picked = None
                    else:
                        o = ev.pick
                        if ev.shift and o:
                            cmdpipe.write("glow %s\n" % o.name)
                            waypoints.append(o.name)
                        elif ev.ctrl and o:
                            cmdpipe.write("glow %s\n"% o.name)
                            cmdpipe.write("traceroute %s %s\n"% (" ".join(waypoints),o.name))
                            waypoints=[]
#                            cmdpipe.write("scan %s\n" % o.name)
#                            cmdpipe.write("reverse_dns %s\n" % o.name)
                        else:
                            if hasattr(o, "action"):
                                o.action()
                if ev.drag == "left":
                    if ev.pick and ev.pick.pickable:
                        picked = ev.pick

    
            if picked:
                picked.PV = picked.pos
                picked.pos=scene.mouse.pos
                picked.V = vector()
                the_world.coords[picked.num]=picked.pos.astuple()
                the_world.speeds[picked.num]=(0.0,0.0,0.0)
                the_world.cinematic_change = 1


            the_world.update(DT)
            the_world.update_edges()
                

        except:
            log.exception("Catched exception in the cinematic thread!")


def mainloop(input, probe, dbq, stereo=None, blind=False):
    feedback = input.get_pipe()

    if not blind:
        scene.title="IPv6 backbone explorer"
        scene.exit=0
        scene.exit_on_close(0)
        if stereo is not None:
            scene.stereo = stereo
            scene.stereodepth = 2
    #        scene.width=100
    #        scene.height=50
        scene.visible=1
    glow=GlowingService()

    unk_node = tag_generator("unk%i")

    def feed_route_back(route, fb):
        if not route:
            return
        if not route[0]:
            route[0] = unk_node.next()
        for k in range(len(route)-1):
            if not route[k+1]:
                route[k+1] = unk_node.next()
            n1,n2 = route[k:k+2]
            fb.write("%s %s\n" % (n1,n2))
#            fb.flush()

    def add_properties(ip, **kargs):
        node = the_world[ip]
        node.update_param(**kargs)
        for k in kargs:
            dbq.set_properties(ip, **kargs)

    def feedback_links_thread(links):
        for l in links:
            feedback.write("%s %s\n"% l)
        feedback.flush()

    the_world = World()
    the_world.scene = scene

    input.add_fd(probe)

    if not blind:
        thread.start_new_thread(cinematic_thread, (the_world, input.get_pipe(), ACCEL, DT) )

    starttime=time.time()
    try:
        while 1:
            cmd = input.get_command()
            try:
                if cmd.startswith("dump"):
                    if " " in cmd:
                        fname = cmd.split()[1]
                    else:
                        fname = "/dev/stdout"
                    the_world.dump_to_file(open(fname,"w"))
                    log.info("Dumped world to [%s]" % fname)
                elif cmd.startswith("dot"):
                    if " " in cmd:
                        fname = cmd.split()[1]
                    else:
                        fname = "/dev/stdout"
                    f = open(fname,"w")
                    f.write('graph "rtgraph" { \n')
                    for e in the_world.edges:
                        f.write('\t "%s" -- "%s";\n' % (e.o1.name, e.o2.name))
                    f.write('}\n')
                    f.close()
                    log.info("Wrote dot file to [%s]" % fname)
                elif cmd.startswith("attract"):
                    the_world.attraction = float(cmd.split()[1])
                elif cmd.startswith("repuls"):
                    the_world.repulsion = float(cmd.split()[1])
                elif cmd.startswith("stereodepth"):
                    scene.stereodepth = int(cmd.split()[1])
                elif cmd.startswith("glow"):
                    name = cmd.split()[1]
                    glow.glow_node(the_world[name])
                elif cmd.startswith("tog"):
                    for n in glow:
                        if n and hasattr(n, "action"):
                            n.action()
                elif cmd.startswith("find"):
                    pattern = cmd.split(" ",1)[1].strip()
                    count=0
                    if "=" in pattern:
                        key,val = [x.strip() for x in pattern.split("=")]
                        rval = re.compile(val)
                        for n in the_world.lst:
                            if key in n.param and rval.search(n.param[key]):
                                count += 1
                                glow.glow_node(n)
                    else:
                        rpat = re.compile(pattern)
                        for n in the_world.lst:
                            if rpat.search(n.name):
                                count += 1
                                glow.glow_node(n)
                    log.info("%i nodes found" % count)                        
                elif cmd.startswith("unglow"):
                    for n in the_world.lst:
                        glow.unglow_node(n)
                elif cmd.startswith("exec"):
                    cmd = cmd.split(" ",1)[1]
                    log.info("--- Begin exec [%s]" % cmd) 
                    res=eval(cmd)
                    log.info(res)
                    log.info("--- End exec [%s]" % cmd) 
                elif cmd.startswith("quit"):
                    try:
                        probe.send("quit\n")
                    finally:
                        raise SystemExit
                    
                elif cmd.startswith("check"):
                    for o in the_world.lst:
                        for o2 in o.edges:
                            if o not in o2.edges:
                                print "%s ==> %s \t %s ==> %s"% (o.name,o2.name, id(o),id(o2))
                    log.info("Checked %i nodes, %i edges." % (len(the_world.lst),len(the_world.edges)))

                elif cmd.startswith("load_ipdb"):
                    thread.start_new_thread(feedback_links_thread,(dbq.get_all_links(),))
                elif cmd.startswith("traceroute"):
                    ip = cmd.split()[1]
                    probe.send_with_cb(cmd, lambda route,fb=feedback: feed_route_back(eval(route),fb))
                elif cmd.startswith("UDP_traceroute"):
                    ip = cmd.split()[1]
                    probe.send_with_cb(cmd, lambda route,fb=feedback: feed_route_back(eval(route),fb))

                elif cmd.startswith("scan"):
                    ip = cmd.split()[1]
                    probe.send_with_cb(cmd, lambda x,ip=ip: add_properties(ip,scan=x))
                elif cmd.startswith("test_rh"):
                    ip = cmd.split()[1]
                    probe.send_with_cb(cmd, lambda x,ip=ip: add_properties(ip,test_rh=x))
                elif cmd.startswith("AS"):
                    ip = cmd.split()[1]
                    probe.send_with_cb(cmd, lambda x,ip=ip: add_properties(ip,AS=x))
                elif cmd.startswith("reverse_dns"):
                    ip = cmd.split()[1]
                    probe.send_with_cb(cmd, lambda x,ip=ip: add_properties(ip,reverse_dns=x))
                elif cmd.startswith("niq_name"):
                    ip = cmd.split()[1]
                    probe.send_with_cb(cmd, lambda x,ip=ip: add_properties(ip,niq_name=x))
                elif cmd.startswith("niq_ipv6"):
                    ip = cmd.split()[1]
                    probe.send_with_cb(cmd, lambda x,ip=ip: add_properties(ip,niq_ipv6=x))
                elif cmd.startswith("listen"):
                    port = int(cmd.split()[1])
                    s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    s.bind(("", port))
                    s.listen(1)
                    t,u=s.accept()
                    s.close()
                    input.add_fd(t)
                    log.info("Connection added from %s:%i" % u)
                elif cmd.startswith("answer"):
                    log.debug("<=== "+cmd)
                    probe.parse_answer(cmd)
                elif cmd:
                    nam1,nam2 = [x.strip().lower() for x in  cmd.split()]
                    if nam1 == nam2:
                        log.info("Nodes are identical")
                        continue
            
                    n1 = the_world.get(nam1)
                    n2 = the_world.get(nam2)
                    if n1 is None:
                        n2,n1=n1,n2
                        nam1,nam2=nam2,nam1
            
                    if n1 is None:
                        attr1=dbq.ip2attr(nam1)
                        n1 = Node(nam1,**attr1)
                        the_world.register(n1)
                        if not nam1.startswith("unk"):
                            dbq.add_ip(nam1)
                            for k in EXISTING_PROPERTIES:
                                if k not in attr1:
                                    feedback.write("%s %s\n" % (k, nam1))
                        
                    if n2 is None:
                        attr2=dbq.ip2attr(nam2)
                        n2 = Node(nam2, pos=n1.pos+randvect(), **attr2)
                        the_world.register(n2)
                        if not nam1.startswith("unk"):
                            dbq.add_ip(nam2)
                            for k in EXISTING_PROPERTIES:
                                if k not in attr2:
                                    feedback.write("%s %s\n" % (k, nam2))
            
                    if n1 in n2.edges:
                        log.info("Edge %s-%s alread exists" % (nam1,nam2))
                        continue
            
                    n1.edges[n2] = None
                    n2.edges[n1] = None
                    the_world.add_edge(n1,n2)
                    if not (nam1.startswith("unk") or nam2.startswith("unk")):
                        dbq.add_link(nam1, nam2)
                    the_world.cinematic_change = 1

                    feedback.flush()
                            
                    
                    log.info("Ok for [%s]-[%s]" % (nam1,nam2))
            except KeyboardInterrupt:
                raise
            except SystemExit:
                raise
            except:
                log.exception("")
                    
    except KeyboardInterrupt:
        log.info("Interrupted by user")
    except SystemExit:
        log.info("Interrupted by command")



class BlindNode(Node):
    def __init__(self, name, *args, **kargs):
        self.param = {"radius":1, "pos": randvect(), "color": randcol() }
        self.param.update(kargs)

        self.name = name
        self.edges = {}

    pos = property(lambda x:vector(), lambda x,y: None)
    def action(self):
        pass
    def update(self, dt):
        pass
    def update_label(self):
        pass

    

class BlindEdge(Edge):
    def __init__(self, o1, o2, *args, **kargs):
        param={"radius":0.2, "color":(0.7,0.7,1)}
        param.update(kargs)
        self.o1=o1
        self.o2=o2
    def update(self):
        pass

        

def usage():
    print >>sys.stderr, """ipv6world [-s stereo] [-p probe] [-i ipdb] [-d dbnet] [-b]
    -p <probe>      command to run probe. Communication is done through stdin/stdout
    -i <ipdb>       path to an IP database
    -d <dbnet>      path to a dbnet database
    -b              blind mode. Do not visualize anything
    -s <stereo>     stereo mode (active, passive, crosseyed, redcyan, redblue, yellowblue)"""
    sys.exit(0)

if __name__ == "__main__":
    STEREO=None
    DBNET="dbnet.db"
    IPDB="ipdb.db"
    BLIND=False
    PROBECMD=["sudo", "./ipv6_probe.py"]
    try:
        opts = getopt.getopt(sys.argv[1:],"hs:d:i:p:b")

        for opt,optarg in opts[0]:
            if opt == "-h":
                usage()
            elif opt == "-s":
                STEREO = optarg
            elif opt == "-d":
                DBNET = optarg
            elif opt == "-i":
                IPDB = optarg
            elif opt == "-p":
                PROBECMD = ["sh","-c",optarg]
            elif opt == "-b":
                BLIND = True
    except getopt.GetoptError,msg:
        log.exception(msg)
        sys.exit(-1)


#    s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#    s.bind(("127.0.0.1",1235))
#    s.listen(5)
#    t,u=s.accept()


    probe = Probe(PROBECMD)
    dbq = DBquery(DBNET,IPDB)

    if BLIND:
        Node=BlindNode
        Edge=BlindEdge
        import rtgraph3d
        rtgraph3d.Edge=BlindEdge
        

    mainloop(MultiInput(sys.stdin), probe, dbq, stereo=STEREO, blind=BLIND)
